@@ -12,13 +12,18 @@ class AuditlogMiddleware:
     def __init__(self, get_response=None):
         self.get_response = get_response
 
-    def __call__(self, request):
-
-        if request.META.get("HTTP_X_FORWARDED_FOR"):
+    @staticmethod
+    def _get_remote_addr(request):
+        if request.headers.get("X-Forwarded-For"):
             # In case of proxy, set 'original' address
-            remote_addr = request.META.get("HTTP_X_FORWARDED_FOR").split(",")[0]
+            remote_addr = request.headers.get("X-Forwarded-For").split(",")[0]
+            # Remove port number from remote_addr
+            return remote_addr.split(":")[0]
         else:
-            remote_addr = request.META.get("REMOTE_ADDR")
+            return request.META.get("REMOTE_ADDR")
+
+    def __call__(self, request):
+        remote_addr = self._get_remote_addr(request)
 
         if hasattr(request, "user") and request.user.is_authenticated:
             context = set_actor(actor=request.user, remote_addr=remote_addr)
