@@ -2,21 +2,27 @@ import json
 
 from django import urls as urlresolvers
 from django.conf import settings
+from django.contrib import admin
+from django.core.exceptions import FieldDoesNotExist
+from django.forms.utils import pretty_name
 from django.urls.exceptions import NoReverseMatch
 from django.utils.html import format_html, format_html_join
 from django.utils.safestring import mark_safe
+from django.utils.timezone import localtime
 
 from auditlog.models import LogEntry
+from auditlog.registry import auditlog
+from auditlog.signals import accessed
 
 MAX = 75
 
 
 class LogEntryAdminMixin:
+    @admin.display(description="Created")
     def created(self, obj):
-        return obj.timestamp.strftime("%Y-%m-%d %H:%M:%S")
-
-    created.short_description = "Created"
+        return localtime(obj.timestamp)
 
+    @admin.display(description="User")
     def user_url(self, obj):
         if obj.actor:
             app_label, model = settings.AUTH_USER_MODEL.split(".")
@@ -29,8 +35,7 @@ class LogEntryAdminMixin:
 
         return "system"
 
-    user_url.short_description = "User"
-
+    @admin.display(description="Resource")
     def resource_url(self, obj):
         app_label, model = obj.content_type.app_label, obj.content_type.model
         viewname = f"admin:{app_label}_{model}_change"
@@ -44,10 +49,9 @@ class LogEntryAdminMixin:
                 '<a href="{}">{} - {}</a>', link, obj.content_type, obj.object_repr
             )
 
-    resource_url.short_description = "Resource"
-
+    @admin.display(description="Changes")
     def msg_short(self, obj):
-        if obj.action == LogEntry.Action.DELETE:
+        if obj.action in [LogEntry.Action.DELETE, LogEntry.Action.ACCESS]:
             return ""  # delete
         changes = json.loads(obj.changes)
         s = "" if len(changes) == 1 else "s"
@@ -57,8 +61,7 @@ class LogEntryAdminMixin:
             fields = fields[:i] + " .."
         return "%d change%s: %s" % (len(changes), s, fields)
 
-    msg_short.short_description = "Changes"
-
+    @admin.display(description="Changes")
     def msg(self, obj):
         changes = json.loads(obj.changes)
 
@@ -80,7 +83,9 @@ class LogEntryAdminMixin:
             msg.append("<table>")
             msg.append(self._format_header("#", "Field", "From", "To"))
             for i, (field, change) in enumerate(sorted(atom_changes.items()), 1):
-                value = [i, field] + (["***", "***"] if field == "password" else change)
+                value = [i, self.field_verbose_name(obj, field)] + (
+                    ["***", "***"] if field == "password" else change
+                )
                 msg.append(self._format_line(*value))
             msg.append("</table>")
 
@@ -98,7 +103,7 @@ class LogEntryAdminMixin:
                     format_html(
                         "<tr><td>{}</td><td>{}</td><td>{}</td><td>{}</td></tr>",
                         i,
-                        field,
+                        self.field_verbose_name(obj, field),
                         change["operation"],
                         change_html,
                     )
@@ -108,8 +113,6 @@ class LogEntryAdminMixin:
 
         return mark_safe("".join(msg))
 
-    msg.short_description = "Changes"
-
     def _format_header(self, *labels):
         return format_html(
             "".join(["<tr>", "<th>{}</th>" * len(labels), "</tr>"]), *labels
@@ -119,3 +122,26 @@ class LogEntryAdminMixin:
         return format_html(
             "".join(["<tr>", "<td>{}</td>" * len(values), "</tr>"]), *values
         )
+
+    def field_verbose_name(self, obj, field_name: str):
+        model = obj.content_type.model_class()
+        try:
+            model_fields = auditlog.get_model_fields(model._meta.model)
+            mapping_field_name = model_fields["mapping_fields"].get(field_name)
+            if mapping_field_name:
+                return mapping_field_name
+        except KeyError:
+            # Model definition in auditlog was probably removed
+            pass
+        try:
+            field = model._meta.get_field(field_name)
+            return pretty_name(getattr(field, "verbose_name", field_name))
+        except FieldDoesNotExist:
+            return pretty_name(field_name)
+
+
+class LogAccessMixin:
+    def render_to_response(self, context, **response_kwargs):
+        obj = self.get_object()
+        accessed.send(obj.__class__, instance=obj)
+        return super().render_to_response(context, **response_kwargs)
